#!/usr/bin/perl

#	VERSION 1.1
#
#	dnsenum.pl: multithread script to enumerate information on a domain.
#		(nslookup, zonetransfer, googlescraping, domain brute force, whois ip, reverse looup)
#	
#	1) Get the host's addresse.
#	2) Get the nameservers (threaded).
#	3) get the MX record (threaded).
#	4) Perform axfr queries on nameservers (threaded).
#	5) Get extra names via google scraping.
#	6) Brute force subdomains from file (threaded).
#	7) Calculate C class domain network ranges and perform whois queries on them (threaded).
#	8) Perform reverse lookups on netranges ( C class or/and whois netranges) (threaded).
#	9) Write to domain_ips.txt file ip-blocks ; results are taken from the reverse lookup.
#
#	See README file.
#
#	TODO: add xml output support.
#
#
#	use with permission.
#
# 	By Filip Waeytens - august 2003 - Filip_Waeytens-at-yahoo.com
# 	threads and extensions by tix - january 2008 - tixxdz-at-gmail.com  
#
#	Special thanks to perl and module developpers.
#

use strict;
use warnings;
use Config;
use Getopt::Long;
use Net::IP;
use Net::DNS;
use Net::Netmask;
use Net::Whois::IP qw(whoisip_query);
use HTML::Parser;
use WWW::Mechanize;

my (%allsubs, %filesubs, %googlesubs);
my (%zonesubs, %reversesubs, %netranges, %recursubs);
my (@nameservers, @mxservers, @results, @ipblocks);
my ($help, $verbose, $whois, $threads, $scrap, $enum);
my ($update, $extend, $ua, $ug, $ur, $uz, $recur);
my ($exp, $subfile, $recursion, $table, $fileips);
my ($domain, $perlthreads, $dnsfile, $noreverse);
my ($timeout, $delay, $pages, $ipcount, $ipvalid) = (10,3,20,0,0); 

# load threads modules
BEGIN {

	if ($Config{useithreads}) {
		$perlthreads = 1;

		require threads;
		require threads::shared;
		require Thread::Queue;

		import threads;
		import threads::shared;
		import Thread::Queue;
	}
	else {
		$perlthreads = 0;
	}
}
 
GetOptions (	'enum'		=>	\$enum,
		'd|delay=i'	=>	\$delay,
		'e|exclude=s'	=>	\$exp,
		'f|file=s'	=>	\$dnsfile,
		'h|help' 	=>	\$help,
		'noreverse'	=>	\$noreverse,
		'p|pages=i'	=>	\$pages,
		'r|recursion'	=>	\$recursion,
		's|scrap=i'	=>	\$scrap,
		'subfile=s'	=>	\$subfile,
		'threads=i'	=>	\$threads,
		't|timeout=i'	=>	\$timeout,
		'u|update=s'	=>	\$update,
		'v|verbose'	=>	\$verbose,
		'w|whois'	=>	\$whois);

usage() if ( $help || !$ARGV[0] );

#options --threads 5 -s 20 -w 
if ($enum) {
	$threads = 5;
	$scrap = 20;
	$whois = 1; 
}

if ($threads && !$perlthreads) {
	print STDERR "Error: your perl version doesn't support ithreads (useithreads)\n";
	exit(1);
}

$domain = lc($ARGV[0]);
$fileips = $domain . "_ips.txt";

#don't test all invalid args (README file)
$threads = 5 if ($threads && $threads <= 0);
$scrap = 10 if ($scrap && $scrap <= 0);

$update = undef if ($update && !$dnsfile);

if ($update) {
	if ($update eq 'g') { $ug = 1; }
	elsif ($update eq 'r') { $ur = 1; }
	elsif ($update eq 'z') { $uz = 1; }
	else { $ua = 1; }  
}

#share vars and populate them with data
if ($threads) {

	#to handle different ips of the domain
	share(@results);

	#number of ips that will be queried in reverse lookup
	share($ipcount);

	#number of valid ips (taken from reverse lookup)
	share($ipvalid); 

	#will contain subdomains found in the zonetransfer results
	share(%zonesubs);

	#will contain subdomains discovred from reverse lookups
	share(%reversesubs) if ($ur);

	#will contain all valid subdomains 
	share(%allsubs);

	#recursion valid subdomains
	share(%recursubs) if ($recursion);

	#to save whois netblocks and it will be used in Net::Netmask functions
	share($table);

	#whois and reverse lookup results
	share(%netranges);

}

print STDOUT "\nChecking " , $domain ,"\n";

# (1) get the host's addresses
print STDOUT "\nHost's addresses: \n";
my $res = Net::DNS::Resolver->new;
$res->udp_timeout($timeout);
my $packet = $res->query($domain, "A");
if ($packet) {
	foreach my $rr (grep { $_->type eq 'A' } $packet->answer) {
		print STDOUT " ", $rr->string , "\n";	
		push @results, $rr->address
			if ($rr->name =~ /$domain/);
	}
}
elsif ($verbose) {
	warn " ", $domain ," A query failed: ", $res->errorstring, "\n";
}

# (2) get the namservers for the domain
print STDOUT "\nNameservers: \n";
$packet = $res->query($domain, "NS");
if ($packet) {

	foreach my $rr (grep { $_->type eq 'NS' } $packet->answer) {
		push @nameservers, $rr->nsdname;
	}

	unless (scalar @nameservers) {
		print STDERR " Error: can't continue no NS record for " , $domain ,"\n";
		exit(1);	
	}

	#read the A record from the additional section
	if ($packet->header->arcount) {
		my @remain = additionalrecord($packet,@nameservers);
		#do nslookup on nameservers that are not present in the additional section
		launchqueries(\&nslookup,@remain) if (scalar @remain); 
	}
	#perform the nslookup on the nameservers
	else {
		launchqueries(\&nslookup,@nameservers);
	}
}
#exit if there is no NS record. 
else {
	warn " ", $domain ," NS record query failed: ", $res->errorstring, "\n";
	exit(1);
}

# (3) get the MX record
print STDOUT "\nMX record: \n";
$packet = $res->query($domain, "MX");
if ($packet) {

	foreach my $rr (grep { $_->type eq 'MX' } $packet->answer) {
		#push mx servers in the @mxservers array (to do nslookup)
		push @mxservers,$rr->exchange; 
	}
	#read the A record directly from the additional section
	if (scalar @mxservers) {
		if ($packet->header->arcount) {
			my @remain = additionalrecord($packet,@mxservers);
			#do nslookup on remaining mx records entries
			launchqueries(\&nslookup,@remain) if (scalar @remain);
		}
		else {
			launchqueries(\&nslookup,@mxservers);	
		}
	}
} 
elsif ($verbose) {
	warn " ", $domain ," MX record query failed: ", $res->errorstring, "\n";
}

#this will save all valid subdomains in %allsubs
#used in recursion, updating brute force file and writing results to a new file
$extend = 1 if ($ua || $subfile || $recursion);

# (4) perform zonetransfers on nameservers 
print STDOUT "\nTrying Zonetransfers: \n";
launchqueries(\&zonetransfer,@nameservers);

# wildcards test
my $wildcards = "pseudorandabcdefuvwxyz0123456789.$domain";
print STDOUT "\nWildcards test: " if ($verbose);
if ($res->query($wildcards)) {
	print STDOUT "\nWildcards detected, all subdomains will point to the same ip address, bay.\n";
	exit(1);
}
elsif ($verbose) {
	print STDOUT "good\n";
}

# (5) scrap additional names from google search and do nslookup on them
if ($scrap) {
	my @tmp = googlescraping();
	launchqueries(\&nslookup,@tmp) if (scalar @tmp);
}

#exit if the brute force file is not specified
unless ($dnsfile) {
	print STDOUT "\nbrute force file not specified, bay.\n";
	exit(1);
}

# (6) brute force subdomains from a file
print STDOUT "\nBrute forcing names from ", $dnsfile ,":\n";
bruteforce("f");

#updating dnsfile
if ($update) {

	#update dnsfile with zonetransfer subdomains
	if ($uz) { 
		writetofile($dnsfile,sort keys %zonesubs) if (%zonesubs);

		#we don't need them anymore
		undef %filesubs;
		undef $update;
	}
	#update dsnfile with google scraping names
	elsif ($ug) {
		writetofile($dnsfile,sort keys %googlesubs) if (%googlesubs);
		
		undef %filesubs;
		undef $update;
	}
}

#free memory
undef %zonesubs if (%zonesubs);
undef %googlesubs if (%googlesubs);

if ($extend) {

	#launch recursion
	if ($recursion) {
		if (%allsubs) {
			print STDOUT "\nPerforming recursion on valid subdomains:\n";
			my @tmp = sort keys %allsubs;
			@tmp = map { $_ .= ".".$domain } @tmp;

			#signal to nslookup to save all recursive subdomains
			$recur = 1;

			#brute force from a list using recursion
			bruteforce("l",\@tmp);
		}
		else {
			print STDOUT " can't perform recursion no valid subdomains\n";	
		}
	}

	#save nameservers and mx servers subdomains in the %allsubs hash
	if ($ua || $subfile) {
		foreach (grep { s/\.$domain// } @nameservers) {
			$allsubs{$_} = 1 unless ($allsubs{$_});
		}
		if (scalar @mxservers) {
			foreach (grep { s/\.$domain// } @mxservers) {
				$allsubs{$_} = 1 unless ($allsubs{$_});
			}
			undef @mxservers;
		}
		undef @nameservers;
	}
	else {
		#we don't need it anymore
		undef %allsubs;
	}
}

# (7) get domain network ranges from brute force results and whois queries
@results = networkranges(sort_by_ip_address(@results));
undef %netranges;

#check domain_ips.txt file 
unlink $fileips or warn "unlink $fileips error: $!, be sure that the file is empty.\n"
	if (-w $fileips);

# (8) perform reverse lookups on netranges (c class or whois netranges)
unless ($noreverse) {
	print STDOUT "\nPerforming reverse lookup on ", $ipcount ," ip addresses:\n";
	launchqueries(\&reverselookup,@results);
	print STDOUT "\n", $ipvalid ," results out of ", $ipcount ," ip addresses.\n";
}
else {
	#calculate ip blocks and write them to domain_ips.txt 
	finalvalidips(sort_by_ip_address(@results));
}

#update dnsfile
if ($update) {

	#update dnsfile with reverse lookups subdomains
	if (!$noreverse && $ur && %reversesubs) {
		writetofile($dnsfile,sort keys %reversesubs);
		undef %reversesubs;
	}
	#update with all subdomains (zonetransfer, google scraping, reverse lookup)
	elsif ($ua) {
		writetofile($dnsfile,sort keys %allsubs);
	}
	undef %filesubs;
}

#writing all valid subdomains to a new file
writetofile($subfile,sort keys %allsubs) if ($subfile);

undef %allsubs;

# (9) show non-contiguous ip blocks 
print STDOUT "\n", $domain ," ip blocks:\n";
foreach (@ipblocks) {
	print STDOUT " ", $_ ,"\n";
}

print STDOUT "\ndone.\n";

exit(0);

#--------------------------------------------------

#subroutine that will launch different queries (nslookp, zonetransfer, whoisip, reverselookup)
sub launchqueries {

	my $querytype = shift;

	if ($threads) {

		if ($querytype != \&reverselookup) {
			my $stream = new Thread::Queue;
			$stream->enqueue(@_);
			my $thr = $threads;  #don't create unused threads
			$thr = scalar @_ if $threads > scalar @_;
		
			for (1 .. $thr) {
				threads->new($querytype,\$stream);	
			}
			#wait all threads
			foreach (threads->list) {
				$_->join if ($_->tid && !threads::equal($_, threads->self)); 
			}	
		}
		else {

			#call reverselookup multiple times to handle multiple netranges 
			#for optimization
			foreach (@_) {
				my $block = new Net::Netmask($_);
				my $stream = new Thread::Queue;
				$stream->enqueue($block->enumerate);

				for (1 .. $threads) {
					threads->new($querytype,\$stream);
				}
				#wait all threads
				foreach (threads->list) {
					$_->join if ($_->tid && !threads::equal($_, threads->self)); 
				}

				#calculate final valid ip blocks for each netrange
				if (%netranges) {
					finalvalidips(sort_by_ip_address(keys %netranges));
					undef %netranges; 
				}
			}
		}
	}
	else {
		if ($querytype != \&reverselookup) {
			foreach (@_) {
				&$querytype($_);
			}
		}
		else {
			foreach (@_) {
				my $block = new Net::Netmask($_);
				&$querytype($block->enumerate);
				if (%netranges) {
					finalvalidips(sort_by_ip_address(keys %netranges));
					undef %netranges;
				}
			}
		}
	}
}

#subroutine to perform reverse lookups
sub reverselookup {

	my $stream = shift if ($threads);

	my $res = Net::DNS::Resolver->new;
	$res->udp_timeout($timeout);
	$res->nameservers(@nameservers);

	while (my $ip = $threads ? $$stream->dequeue_nb : shift) { 
		
		my $query = $res->query($ip);
		if ($query) {
			foreach my $rr ($query->answer) {
				
				#exclude non PTR types answers or unwanted hostnames
				next if ($rr->type ne "PTR" || ($exp && $rr->ptrdname =~ /$exp/));
				 
				if ($rr->ptrdname =~ /(.*)(\.$domain.*)/i) {

					$allsubs{$1} = 1 if ($extend && !$allsubs{$1});
					$reversesubs{$1} = 1 if ($ur && !$reversesubs{$1});

					#to calculate last valid ip blocks
					unless ($netranges{$ip}) {
						$netranges{$ip} = 1;
						$ipvalid++;
					}
					print STDOUT " ", $rr->string ,"\n";
				}
				#show all answers even if the hostname don't match the domain
				elsif ($verbose) {
					print STDOUT " ", $rr->string ,"  \n" 
				}	
			}
		}
		#this part is just to check progress
		elsif ($verbose) {
		      print STDOUT " ", $ip ,"    ...\n";
		}
	}
}

#subroutine for nslookups
sub nslookup {

	my $stream = shift if ($threads);

	my $res = Net::DNS::Resolver->new;
	$res->udp_timeout($timeout);

	#don't append items from the resolver search list
	$res->dnsrch(0);

	while (my $host = $threads ? $$stream->dequeue_nb : shift) {

		my $query = $res->search($host);
		if ($query) {
			foreach my $rr ($query->answer) {
				print STDOUT " ", $rr->string , "\n";

				#check if it match the domain
				if ($rr->name =~ /(.*)(\.$domain.*)/i) {
	
					#save valid subdomains
					if ($extend && !$allsubs{$1}) {
						$allsubs{$1} = 1;
						
						#for recursion
						$recursubs{$1} = 1 if ($recur);
					}

					#save ip address 
					push @results, $rr->address if ($rr->type eq 'A');

				}
 			}
		}
		elsif ($verbose) {
			warn " ", $host ," A record query failed: ", $res->errorstring, "\n";
		}
	}
}

#subroutine for zonetransfers
sub zonetransfer {

	my $stream = shift if ($threads);

	my $res = Net::DNS::Resolver->new;
	$res->tcp_timeout($timeout);
	$res->udp_timeout($timeout);

	while (my $ns = $threads ? $$stream->dequeue_nb : shift) { 
	  	$res->nameservers($ns);
		
		my @zone = $res->axfr($domain);
		print STDOUT "\n trying zonetransfer for ", $domain ," on ", $ns ," ... \n";
		if (@zone) {
			foreach my $rr (@zone) {
				#print all results
				print STDOUT " ", $rr->string , "\n";

				#save the subdomains and the ip addresses
				if ( $rr->name =~ /(.*)(\.$domain.*)/i) {
					
					$allsubs{$1} = 1 if ($extend && !$allsubs{$1});

					$zonesubs{$1} = 1 unless ($zonesubs{$1});
					
					push @results, $rr->address if ($rr->type eq 'A');
				}
			}
		}	
		elsif ($verbose) {
			warn "  AXFR record query failed: ", $res->errorstring, "\n";
		}
	}
}

#subroutine for scraping domains from google 
sub googlescraping {

	my ($count, $mypage, $query) = (0,1,"allinurl: -www site:".$domain);
	my ($response, $browser, $parser, $nextpage, @subdomains);
	print STDOUT "\nScraping ", $domain ," subdomains from google ... \n";
	
	#on errors the mech object will call die
	$browser = WWW::Mechanize->new(	autocheck	=>	1,
					stack_depth 	=> 	1,
					cookie_jar 	=> 	undef);

	my @agents = $browser->known_agent_aliases();
	my $agent = $agents[rand(scalar @agents)];
	$browser->agent_alias($agent);
	$browser->timeout($timeout);

	#get the first page
	$response = $browser->get("http://www.google.com/ncr");
	
	my $form = $browser->form_number(1);

	return unless ($form);

	#submit the form f
	$form->find_input("q")->value($query);
	$response = $browser->submit_form(	form_number	=>	1,
						form_name	=>	"f");
	do {
		$nextpage = undef;
		print STDOUT " Processing google search page: ", $mypage ,"\n" if ($verbose);

		$parser = HTML::Parser->new(
				api_version => 3,	
				start_h	=> [sub {
						my $tag = shift;

						#end of parsing (we have enough dns names) 
						$parser->eof unless ($count < $scrap);

						return unless $tag->{href};
							
						#subdomains checks
						if ($tag->{href} =~ /(^http:\/\/)(.*)(\.$domain\/$)/) {

							#print results
							print STDOUT "  name:  ", $2 ,"\n" if ($verbose);

							$googlesubs{$2} = 1 unless ($googlesubs{$2});

							#don't save it if it has been already founded in 
							#zonetransfer result.
							push @subdomains, $2.".".$domain unless ($zonesubs{$2});

							#count all google results even it has been 
							#found by zonetransfer
							$count++;

						}

						#the next page
						elsif ($tag->{href} =~ /^\/search\?q=.*$domain.*start=$mypage\d.*/) {
							$nextpage = $tag->{href} unless ($nextpage); 
						}

						},'attr'] 
					);

		$parser->parse($response->decoded_content);

		if ($nextpage) {	
			$response = $browser->get($nextpage);
			$mypage++;
		}

	} while ($count < $scrap && $mypage <= $pages && $nextpage);

	print STDOUT "\ngoogle results: ", $count ,"\n";

	if ($count) {
	 	return @subdomains;
	}
	else {
		print STDOUT "perhaps google is blocking our queries.\n"; 
		return;
	}
}

#subroutine to query a whois server for an ip address to get the correct netrange
sub whoisip {

	my $stream = shift if ($threads);

	while (my $ip = $threads ? $$stream->dequeue_nb : shift) {
		my ($inetnum, $block);

		#search in the network blocks table to find if any of them contains the IP address
		next if (findAllNetblock($ip,$table));
		
		#this is very useful on whois servers that limit the number of connections
		sleep(rand($delay));
	
		#catch different exceptions (on exceptions netrange will be a c class /24)
		eval {
			my $response = whoisip_query($ip);
			foreach (keys %{$response}) {

				next if ($_ !~ /^(inetnum|netrange)$/i);

				$inetnum = $response->{$_};

				#handle all whois netrange format
				if ($inetnum =~ /([\d.\.]+)\/(\d{1,2})/) {
					#whois.lacnic.net format
					$block = new2 Net::Netmask ("$1/$2");	
				}
				else {
					$inetnum =~ s/\s//g;
					$block = new2 Net::Netmask ($inetnum);
				}	

				if ($block) {

					#this is useful when threads are enabled
					#to eliminate processing same netranges 
					next if ($threads && $netranges{$block->desc});

					$block->storeNetblock($table);

					#this is a simple workaround to save netblocks in a hash 
					#because we will lost all data in $table after threads exit 
					#(see Net::Netmask and threads::shared and threads safe)
					#i am soure that there is a beter solution
					#please excuse my ignorance
					$netranges{$block->desc} = 1;

					$ipcount += $block->size();
					printf STDOUT " whois ip result:   %-15s    ->      %s\n", 
							$ip, $block->desc;
				}
				else {
					print STDERR " Netmask error: ", $Net::Netmask::error ,"\n" 
						if ($verbose);
					$inetnum = undef;
				}
			}
		};
		if ($@) {
			#catch any invalid results 
			#assume that the network range is a c class
			print STDERR " Error: ", $@ if ($verbose);
			$inetnum = undef;
		}

		#can't get the netrange form the whois server
		#so we assume that is a c class range /24
		unless (defined $inetnum) {
			$block = new Net::Netmask ("$ip/24");
			next if ($threads && $netranges{$block->desc});
				
			$block->storeNetblock($table);
			$netranges{$block->desc} = 1;
			$ipcount += 256;	
			printf STDOUT " c class default:   %-15s    ->      %s  "
					."    (whois netrange operation failed)\n",
					$ip, $block->desc;
		}
	}	
}

#subroutine that reads the A record from the additional section
sub additionalrecord {

	my ($packet,@servers,@remain,%seen) = @_;
	
	foreach my $rr (grep { $_->type eq 'A' } $packet->additional) {
		foreach (@servers) {

			next if ($rr->name ne $_);

			$seen{$rr->name} = 1;
			print STDOUT " ", $rr->string , "\n";	
			push @results, $rr->address
				if ($rr->name =~ /$domain/);

		} 
	}
	#get the nameservers that have not been found in the additional section
	unless ( keys %seen == @servers ) {
		foreach (@servers) {
			push @remain,$_ unless ($seen{$_});
		}	
	}
	return @remain;	
}

#subroutine to brute force subdomains
sub bruteforce {

	my @words;
	#recursion on valid subdomains brute force from a list
	if (shift eq "l") {
		my ($hosts, @tmp);

		RECURSION:
		$hosts = shift;
		#get subdomains to perform brute force
		my @subdomains = sort keys %allsubs;
		push @subdomains, grep { not $allsubs{$_} } sort keys %filesubs 
			if (%filesubs);
		
		foreach my $host (@$hosts) {
			foreach (@subdomains) {
				push @words, "$_.$host";
			}
			print STDOUT "\n Recursion on ", $host ," ...\n";
			launchqueries(\&nslookup,@words);
			undef @words; #simply to free up some memory	
		}

		return unless (%recursubs);	

		@tmp = map { $_ .= ".".$domain } sort keys %recursubs;
		undef %recursubs;
		@_ = \@tmp; 
		goto RECURSION;
	}
	#brute force subdomains from dnsfile
	else {
		open(INPUT, "<" . $dnsfile) || die "could not open subdomain $dnsfile file: $!\n";
		while (<INPUT>) {
			chomp;	
			$filesubs{$_} = 1 if ($update || $recursion);

			#select all subdomains that have not been listed 
			push @words, $_.".".$domain 
				unless ($allsubs{$_} || $zonesubs{$_} || $googlesubs{$_});	
		}	
		close(INPUT);
	
		unless (scalar @words) {
			#file empty or all subdomains already listed
			print STDOUT "No subdomains.\n";
		}
		else {
			#brute force all names
			launchqueries(\&nslookup,@words);
		}
	}
}

#subroutine to write valid subdomains to files
sub writetofile {
	my $file = shift;
	print STDOUT "\nWriting to ", $file ," valid subdomains ... \n";

	#updating dnsfile with valid subdomains
	if ($file eq $dnsfile) {
		unless (open(OUTPUT, ">>" . $file)) {
			warn "could not open subdomain $file file: $!\n";
			return;
		}
		else {
			foreach (@_) {
				print OUTPUT $_ ,"\n" unless ($filesubs{$_});
			}
			close(OUTPUT);
		}
	}

	#writing valid subdomains to a new file
	else {
		unless (open(OUTPUT, ">" . $file)) {
			warn "could not open subdomain $file file: $!\n";
			return;
		}
		else {
			foreach (@_) {
				print OUTPUT $_ ,"\n";
			}
			close(OUTPUT);
		}
	}
}

#subroutine to get domain netranges
sub networkranges {

	my (@tmp, %seen, %ips);
	
	#get uniq ips and uniq c class netranges
	foreach (@_) {
		my @octets = split /\./, $_;
		
		#eliminate private ips
		next if ($octets[0] == 10 
			|| $octets[0] == 127 
			|| ($octets[0] == 169 && $octets[1] == 254) 
			|| ($octets[0] == 172 && ($octets[1] > 15 && $octets[1] < 32 ))
			|| ($octets[0] == 192 && $octets[1] == 168 ));

		#to get uniq c class netranges
		my $net = "$octets[0].$octets[1].$octets[2]";
		unless ($seen{$net}) {
			$seen{$net} = 1;
			push @tmp, "$net.0";
		}

		#uniq ips 
		$ips{$_} = 1 unless ($ips{$_});
		
	}	

	#launch whois queries on ips to get the correct netrange
	if ($whois) {
		print STDOUT "\nLunching whois queries ...\n";

		#shutdown warns the whoisip will catch exceptions with eval
		$SIG{__WARN__} = sub {} ;
		launchqueries(\&whoisip,@tmp);
		$SIG{__WARN__} = 'DEFAULT';
		
		print STDOUT "\n", $domain ," whois netranges:\n";
		foreach (keys %netranges) {
			print STDOUT " ", $_ ,"\n";
		} 
		@tmp = keys %netranges;
	}
	#default c class netrange
	else {
		print STDOUT "\n", $domain ," c class netranges:\n";
		foreach (@tmp) {
			$_ .= "/24";
			$ipcount += 256;
			print STDOUT " ", $_ ,"\n";
		}
	}
	#return c class or whois netrnages
	unless ($noreverse) {
		return @tmp;
	}
	#return uniq ips 
	else {
		return keys %ips;
	}	
}

#subroutine that calculate non-contiguous ip blocks
sub finalvalidips {

	my ($firstip,$lastip) = shift;
	open(OUTPUT, ">>" . $fileips) || die "could not open file $fileips $!\n";

	if (scalar @_) {
		my $broadcast = $_[$#_];
		my $tmp = new Net::IP("$firstip - $broadcast");
		foreach (@_) {
			#increment the previous tmp ip address to compare it with the current 
			#ip address taken from the array 
			++$tmp;

			if ($broadcast ne $_) {
			
				#this ip belongs to the current netrange
				if ($tmp->ip() eq $_) {
				$lastip = $_;
				} 
				#new netrange
				else {
					#save the previous netrange
					if ($lastip) {
						my @list = getcidrs($firstip, $lastip);
						push @ipblocks, @list;
						foreach (@list) {
							print OUTPUT $_ ,"\n";
						}	
					}
					#save the previous ip (netrange with a single ip)
					else {
						push @ipblocks, $firstip."/32";
						print OUTPUT $firstip ,"/32\n";	
					}
					#update data
					$firstip = $_;
					$lastip = undef;
					$tmp = new Net::IP($firstip." - ".$broadcast);
				}
			}
			#we have reached the last valid ip address in the network range
			else {
				#this ip belongs to the current range
				if ($tmp->ip() eq $broadcast) {
					my @list = getcidrs($firstip, $broadcast);;
					push @ipblocks, @list;
					foreach (@list) {
						print OUTPUT $_ ,"\n";
					}
				}
				#this ip is the start of a new range 
				else {
					#save the previous netrange
					if ($lastip) {
						my @list = getcidrs($firstip, $lastip);
						push @ipblocks, @list;
						foreach (@list) {
							print OUTPUT $_ ,"\n";
						}
					}
					#save the previous ip (netrange with a single ip)
					else {
						push @ipblocks, $firstip."/32";
						print OUTPUT $firstip ,"/32\n";
					}	
					#save the current new ip
					push @ipblocks, $broadcast."/32";
					print OUTPUT $broadcast ,"/32\n";
				}	
			}
		}
	}
	#one single ip address
	else {
		push @ipblocks, $firstip."/32";
		print OUTPUT $firstip ,"/32\n";
	}
	close(OUTPUT);
}

#subroutine that returns a list of netranges with cidrs notation
sub getcidrs {

	my @ranges;
	foreach (range2cidrlist(shift, shift)) {
		push @ranges, $_->desc;	
	}
	return @ranges;
}

#the usage subroutine
sub usage {
	print STDERR 
qq{Usage: $0 [options] <domain> 
 Enumerate information on a domain and discover non-contiguous ip blocks.
 [options]:
 Note: the brute force -f switch must be specified to be able to continue the process execution.

 General options:				
	--enum				Shortcut option equivalent to --threads 5 -s 20 -w.
  -h,	--help				Print this help message.
	--noreverse			Skip the reverse lookup operations.
	--subfile	file		Write all valid subdomains to this file. (may consume lot of memory).
  -t,	--timeout	value		The tcp and udp timeout values in seconds (default: 10s).
	--threads	value		The number of threads that will perform different queries.
  -v,	--verbose			Be verbose (show all the progress and all the error messages).

 Google scraping options:
  -p,	--pages		value		The number of google search pages to process when scraping names
						(default: 20 pages), the -s switch must be specified.
  -s,	--scrap		value		The maximum number of subdomains that will be scraped from google.

 Brute force options:
  -f,	--file		file		Read subdomains from this file to perform brute force.
  -u,	--update	a|g|r|z		Update the file specified with the -f switch with vaild subdomains.
			a (all)		Update using all results, invalid values will be a.
			g		Update using only google scraping results.
			r		Update using only reverse lookup results. (may consume lot of memory).
			z		Update using only zonetransfer results. 
  -r,	--recursion			Recursion on subdomains, brute force all discovred subdomains.

 Whois netrange options:
  -d,	--delay		value		The maximum value of seconds to wait between whois queries, the 
						value is defined randomly, (default: 3s).
  -w,	--whois				Perform the whois queries on netranges. 
					**Warning**: this can generate very large netranges and it will take
						     lot of time to enumerate them.

 Reverse lookup options:
  -e,	--exclude	regexp		Exclude PTR records that match the regexp expression from reverse
						lookup results, useful on invalid hostnames.	
};
        exit(1);
}
